//(C) Copyright Gerald Fahrnholz 2009

using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;

namespace AnalyzeTestCoverage
{
    public class XmlAnalyzer
    {
        // Default settings of commandline params
        private string m_fileName = "";
        private bool m_detailed = false;
        private bool m_onlyDisplayHelp = false;
        private int m_ignoreClassesWithFewerLinesThan = 5;

        // Command line syntax
        private static readonly string OPT_HELP_1 = "/?";
        private static readonly string OPT_HELP_2 = "-?";
        private static readonly string OPT_HELP_3 = "/help";
        private static readonly string OPT_HELP_4 = "-help";
        private static readonly string OPT_FILE_NAME = "/file=";
        private static readonly string OPT_INCLUDE = "/include=";
        private static readonly string OPT_EXCLUDE = "/exclude=";
        private static readonly string OPT_DETAILED = "/detailed";
        private static readonly string OPT_MIN_NUM_LINES = "/minNumLines=";

        //---------------------------------------------------------------------

        public XmlAnalyzer()
        {
            m_prefixesToInclude = new System.Collections.ArrayList();
            m_prefixesToExclude = new System.Collections.ArrayList();
        }

        //---------------------------------------------------------------------

        public void InitFromCommandline()
        {
            try
            {
                string[] args = System.Environment.GetCommandLineArgs();
                foreach (string arg in args)
                {
                    if ((arg == OPT_HELP_1) || (arg == OPT_HELP_2) ||
                        (arg == OPT_HELP_3) || (arg == OPT_HELP_4))
                    {
                        m_onlyDisplayHelp = true;
                    }
                    else if (arg.StartsWith(OPT_FILE_NAME))
                    {
                        m_fileName = arg.Substring(OPT_FILE_NAME.Length, arg.Length - OPT_FILE_NAME.Length);
                    }
                    else if (arg.StartsWith(OPT_INCLUDE))
                    {
                        string prefixToInclude = arg.Substring(OPT_INCLUDE.Length, arg.Length - OPT_INCLUDE.Length);
                        prefixToInclude.Trim();
                        if (prefixToInclude.Length > 0)
                        {
                            m_prefixesToInclude.Add(prefixToInclude);
                        }
                    }
                    else if (arg.StartsWith(OPT_EXCLUDE))
                    {
                        string prefixToExclude = arg.Substring(OPT_EXCLUDE.Length, arg.Length - OPT_EXCLUDE.Length);
                        prefixToExclude.Trim();
                        if (prefixToExclude.Length > 0)
                        {
                            m_prefixesToExclude.Add(prefixToExclude);
                        }
                    }
                    else if (arg == OPT_DETAILED)
                    {
                        m_detailed = true;
                    }
                    else if (arg.StartsWith(OPT_MIN_NUM_LINES))
                    {
                        m_ignoreClassesWithFewerLinesThan = int.Parse(
                            arg.Substring(OPT_MIN_NUM_LINES.Length, arg.Length - OPT_MIN_NUM_LINES.Length));
                    }
                    else if (arg.Contains("AnalyzeTestCoverage"))
                    {
                        // do nothing, first argument containing name of executable
                    }
                    else
                    {
                        throw new SystemException("Wrong argument: " + arg);
                    }
                }

                if (m_onlyDisplayHelp)
                {
                    Console.WriteLine("\nAnalyzeTestCoverage, (C) Copyright Gerald Fahrnholz 2009\n");
                    DisplayInfoAndExit(0);
                }
            }
            catch (SystemException ex)
            {
                Console.WriteLine(
                    "\nERROR when parsing command line"
                      + "\n\n" + ex.Message);
                DisplayInfoAndExit(1);
            }
        }

        //---------------------------------------------------------------------

        public void EvaluateResults()
        {
            OpenResultFile();

            XmlNodeList xmlNodeList = m_xmlDoc.SelectNodes("CoverageDSPriv/Module/NamespaceTable/Class");
            int numAvailableClassEntries = xmlNodeList.Count;

            int totalNumAnalyzedClasses = 0;
            int totalNumLines = 0;
            int totalNumLinesCovered = 0;
            int totalNumBlocks = 0;
            int totalNumBlocksCovered = 0;

            foreach (XmlNode node in xmlNodeList)
            {
                XmlNode classNameNode = node.SelectSingleNode("ClassName");
                string className = classNameNode.InnerText;
                className.Trim();
                if (className.Length == 0)
                {
                    className = node.SelectSingleNode("ClassKeyName").InnerText + " (ClassKeyName)";
                }

                if (ClassIsIncluded(className) && !ClassIsExcluded(className))
                {
                    int linesCovered = int.Parse(node.SelectSingleNode("LinesCovered").InnerText);
                    int linesNotCovered = int.Parse(node.SelectSingleNode("LinesNotCovered").InnerText);
                    int linesPartiallyCovered = int.Parse(node.SelectSingleNode("LinesPartiallyCovered").InnerText);
                    linesCovered += linesPartiallyCovered;
                    int lines = linesCovered + linesNotCovered;
                    double percentCoveredLines = 100.0 * ((linesCovered * 1.0) / lines);

                    int blocksCovered = int.Parse(node.SelectSingleNode("BlocksCovered").InnerText);
                    int blocksNotCovered = int.Parse(node.SelectSingleNode("BlocksNotCovered").InnerText);
                    int blocks = blocksCovered + blocksNotCovered;
                    double percentCoveredBlocks = 100.0 * ((blocksCovered * 1.0) / (blocks));

                    if (lines >= m_ignoreClassesWithFewerLinesThan) // ignore small classes (probably only simple data types)
                    {
                        Console.Write(className + ";"
                            + lines + ";"
                            + Math.Round(percentCoveredLines, 0).ToString());
                        if (m_detailed)
                        {
                            Console.Write(";" + blocks + ";" + Math.Round(percentCoveredBlocks, 0).ToString());
                        }
                        Console.WriteLine("");

                        ++totalNumAnalyzedClasses;
                        totalNumLines += lines;
                        totalNumLinesCovered += linesCovered;
                        totalNumBlocks += blocks;
                        totalNumBlocksCovered += blocksCovered;
                    }
                }
            }
            double totalPercentCoveredLines = 100.0 * ((totalNumLinesCovered * 1.0) / totalNumLines);
            double totalPercentCoveredBlocks = 100.0 * ((totalNumBlocksCovered * 1.0) / totalNumBlocks);
            string info = "legend: class / numLines / % covered";
            if (m_detailed)
            {
                info += "/ numBlocks / % covered";
            }
            Console.Write("_Total, analyzed " + totalNumAnalyzedClasses + " of "
                + numAvailableClassEntries + " classes, " + info + ";" + totalNumLines + ";"
                    + Math.Round(totalPercentCoveredLines, 0).ToString());

            if (m_detailed)
            {
                Console.Write(";" + totalNumBlocks + ";"
                    + Math.Round(totalPercentCoveredBlocks, 0).ToString());
            }
        }

        //---------------------------------------------------------------------
        // Private methods
        //---------------------------------------------------------------------

        private void OpenResultFile()
        {
            try
            {
                m_xmlDoc = new XmlDocument();
                m_xmlDoc.Load(m_fileName);
            }
            catch (SystemException ex)
            {
                Console.WriteLine(
                    "\nERROR" +
                    "\nCould not open file >" + m_fileName + "<" +
                    "\nCurrent directory is " + System.Environment.CurrentDirectory);
                WriteExceptionInfo(ex);
                DisplayInfoAndExit(1);
            }
        }

        //---------------------------------------------------------------------

        private bool ClassIsIncluded(string in_className)
        {
            bool retVal = false;
            if (m_prefixesToInclude.Count == 0)
            {
                retVal = true;
            }
            else
            {
                foreach (string prefixToInclude in m_prefixesToInclude)
                {
                    if (in_className.StartsWith(prefixToInclude))
                    {
                        retVal = true;
                        break;
                    }
                }
            }
            return retVal;
        }

        //---------------------------------------------------------------------

        private bool ClassIsExcluded(string in_className)
        {
            bool retVal = false;
            foreach (string prefixToExclude in m_prefixesToExclude)
            {
                if (in_className.StartsWith(prefixToExclude))
                {
                    retVal = true;
                    break;
                }
            }
            return retVal;
        }

        //---------------------------------------------------------------------

        private void DisplayInfoAndExit(int in_exitCode)
        {
            WriteCurSettings();
            WriteCommandlineSyntax();
            System.Environment.Exit(in_exitCode);
        }

        //---------------------------------------------------------------------

        private void WriteCurSettings()
        {
            Console.WriteLine("\n\nCURRENT SETTINGS\n");
            Console.WriteLine("Current directory                     : " + System.Environment.CurrentDirectory);
            Console.WriteLine("File name                             : " + m_fileName);
            Console.WriteLine("Detailed output                       : " + m_detailed);
            Console.WriteLine("Only display help                     : " + m_onlyDisplayHelp);
            Console.WriteLine("Ignore classes with fewer lines than  : " + m_ignoreClassesWithFewerLinesThan);

            Console.Write("Class names to include                :");
            if (m_prefixesToInclude.Count == 0)
            {
                Console.Write(" all (none specified)");
            }
            else
            {
                foreach (string prefixToInclude in m_prefixesToInclude)
                {
                    Console.Write(" " + prefixToInclude);
                }
            }
            Console.WriteLine("");

            Console.Write("Class names to exclude                :");
            if (m_prefixesToExclude.Count == 0)
            {
                Console.Write(" - (none specified)");
            }
            else
            {
                foreach (string prefixToExclude in m_prefixesToExclude)
                {
                    Console.Write(" " + prefixToExclude);
                }
            }
            Console.WriteLine("");
        }

        //---------------------------------------------------------------------

        private void WriteCommandlineSyntax()
        {
            Console.WriteLine("\n\nSYNTAX\n\nAnalyzeTestCoverage /file=<xmlFileName> [/include=<classPrefix>] [/exclude=<classPrefix>] [/minNumLines=<int>] [/detailed] [/?]\n");
            Console.WriteLine("/file=<xmlFileName>\nResult file exported as XML from TestCoverage\n\n");
            Console.WriteLine("/include=<classPrefix>\nAnalyze only outputs for classes starting with the given prefix."
                + "\nYou can specify several prefixes, e.g.: /include=MyClass /include=SomeOtherClass"
                + "\noptional param, default: all classes\n\n");
            Console.WriteLine("/exclude=<classPrefix>\nAnalyze only outputs for classes NOT starting with the given prefix."
                + "\nYou can specify several prefixes, e.g.: /exclude=boost:: /exclude=std::"
                + "\noptional param, default: none\n\n");
            Console.WriteLine("/minNumLines=<int>\nIgnore classes with fewer number of lines."
                + "\noptional param, default: 5 lines\n\n");
            Console.WriteLine("/detailed\nProduce more detailed output (e.g. covered blocks)."
                + "\noptional param, default: false");
        }

        //---------------------------------------------------------------------

        private void WriteExceptionInfo(SystemException ex)
        {
            Console.WriteLine("\n\nEXCEPTION-INFO\n" + ex.Message);
        }

        //---------------------------------------------------------------------
        // Private data
        //---------------------------------------------------------------------

        private XmlDocument m_xmlDoc;
        private System.Collections.ArrayList m_prefixesToInclude;
        private System.Collections.ArrayList m_prefixesToExclude;
    }
}
